'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any

from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.ocel.util import extended_table
from pm4py.objects.ocel.util import ocel_consistency
from enum import Enum
from pm4py.util import exec_utils, constants as pm4_constants, pandas_utils


class Parameters(Enum):
    ENCODING = "encoding"


def apply(file_path: str, objects_path: str = None, parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    """
    Imports an object-centric event log from a CSV file, using Pandas as backend

    Parameters
    -----------------
    file_path
        Path to the object-centric event log
    objects_path
        Optional path to a CSV file containing the objects dataframe
    parameters
        Parameters of the algorithm

    Returns
    ------------------
    ocel
        Object-centric event log
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(Parameters.ENCODING, parameters, pm4_constants.DEFAULT_ENCODING)
    table = pandas_utils.read_csv(file_path, index_col=False, encoding=encoding, dtype=str)

    objects = None
    if objects_path is not None:
        objects = pandas_utils.read_csv(objects_path, index_col=False, encoding=encoding, dtype=str)

    ocel = extended_table.get_ocel_from_extended_table(table, objects, parameters=parameters)
    ocel = ocel_consistency.apply(ocel, parameters=parameters)

    return ocel



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.algo.discovery.heuristics.variants import classic, plusplus
from pm4py.objects.conversion.log import converter as log_conversion
from pm4py.objects.heuristics_net.obj import HeuristicsNet
from pm4py.util import exec_utils, pandas_utils
from typing import Optional, Dict, Any, Union, Tuple
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from pm4py.objects.petri_net.obj import PetriNet, Marking


class Variants(Enum):
    CLASSIC = classic
    PLUSPLUS = plusplus


CLASSIC = Variants.CLASSIC
DEFAULT_VARIANT = CLASSIC

VERSIONS = {CLASSIC}


def apply(log: Union[EventLog, EventStream, pd.DataFrame], parameters: Optional[Dict[Any, Any]] = None, variant=CLASSIC) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using Heuristics Miner

    Parameters
    ------------
    log
        Event log
    parameters
        Possible parameters of the algorithm,
        including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY
            - Parameters.CASE_ID_KEY
            - Parameters.DEPENDENCY_THRESH
            - Parameters.AND_MEASURE_THRESH
            - Parameters.MIN_ACT_COUNT
            - Parameters.MIN_DFG_OCCURRENCES
            - Parameters.DFG_PRE_CLEANING_NOISE_THRESH
            - Parameters.LOOP_LENGTH_TWO_THRESH
    variant
        Variant of the algorithm:
            - Variants.CLASSIC
            - Variants.PLUSPLUS

    Returns
    ------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    if pandas_utils.check_is_pandas_dataframe(log):
        return exec_utils.get_variant(variant).apply_pandas(log, parameters=parameters)

    return exec_utils.get_variant(variant).apply(log_conversion.apply(log, variant=log_conversion.Variants.TO_EVENT_LOG, parameters=parameters),
                                                 parameters=parameters)


def apply_dfg(dfg: Dict[Tuple[str, str], int], activities=None, activities_occurrences=None, start_activities=None, end_activities=None,
              parameters=None, variant=CLASSIC) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using Heuristics Miner

    Parameters
    ------------
    dfg
        Directly-Follows Graph
    activities
        (If provided) list of activities of the log
    activities_occurrences
        (If provided) dictionary of activities occurrences
    start_activities
        (If provided) dictionary of start activities occurrences
    end_activities
        (If provided) dictionary of end activities occurrences
    parameters
        Possible parameters of the algorithm,
        including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY
            - Parameters.CASE_ID_KEY
            - Parameters.DEPENDENCY_THRESH
            - Parameters.AND_MEASURE_THRESH
            - Parameters.MIN_ACT_COUNT
            - Parameters.MIN_DFG_OCCURRENCES
            - Parameters.DFG_PRE_CLEANING_NOISE_THRESH
            - Parameters.LOOP_LENGTH_TWO_THRESH
    variant
        Variant of the algorithm:
            - Variants.CLASSIC

    Returns
    ------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    return exec_utils.get_variant(variant).apply_dfg(dfg, activities=activities,
                                                     activities_occurrences=activities_occurrences,
                                                     start_activities=start_activities, end_activities=end_activities,
                                                     parameters=parameters)


def apply_heu(log: Union[EventLog, EventStream, pd.DataFrame], parameters: Optional[Dict[Any, Any]] = None, variant=CLASSIC) -> HeuristicsNet:
    """
    Discovers an Heuristics Net using Heuristics Miner

    Parameters
    ------------
    log
        Event log
    parameters
        Possible parameters of the algorithm,
        including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY
            - Parameters.CASE_ID_KEY
            - Parameters.DEPENDENCY_THRESH
            - Parameters.AND_MEASURE_THRESH
            - Parameters.MIN_ACT_COUNT
            - Parameters.MIN_DFG_OCCURRENCES
            - Parameters.DFG_PRE_CLEANING_NOISE_THRESH
            - Parameters.LOOP_LENGTH_TWO_THRESH
    variant
        Variant of the algorithm:
            - Variants.CLASSIC

    Returns
    ------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    return exec_utils.get_variant(variant).apply_heu(log_conversion.apply(log, variant=log_conversion.Variants.TO_EVENT_LOG, parameters=parameters), parameters=parameters)


def apply_heu_dfg(dfg: Dict[Tuple[str, str], int], activities=None, activities_occurrences=None, start_activities=None, end_activities=None,
                  parameters=None, variant=CLASSIC) -> HeuristicsNet:
    """
    Discovers an Heuristics Net using Heuristics Miner

    Parameters
    ------------
    dfg
        Directly-Follows Graph
    activities
        (If provided) list of activities of the log
    activities_occurrences
        (If provided) dictionary of activities occurrences
    start_activities
        (If provided) dictionary of start activities occurrences
    end_activities
        (If provided) dictionary of end activities occurrences
    parameters
        Possible parameters of the algorithm,
        including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY
            - Parameters.CASE_ID_KEY
            - Parameters.DEPENDENCY_THRESH
            - Parameters.AND_MEASURE_THRESH
            - Parameters.MIN_ACT_COUNT
            - Parameters.MIN_DFG_OCCURRENCES
            - Parameters.DFG_PRE_CLEANING_NOISE_THRESH
            - Parameters.LOOP_LENGTH_TWO_THRESH
    variant
        Variant of the algorithm:
            - Variants.CLASSIC

    Returns
    ------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    return exec_utils.get_variant(variant).apply_heu_dfg(dfg, activities=activities,
                                                         activities_occurrences=activities_occurrences,
                                                         start_activities=start_activities,
                                                         end_activities=end_activities,
                                                         parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Tuple, List, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL, IMDataStructureDFG
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough
from pm4py.algo.discovery.inductive.fall_through.empty_traces import EmptyTracesUVCL
from pm4py.objects.process_tree.obj import ProcessTree, Operator
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL
from pm4py.objects.dfg.obj import DFG
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG


class FlowerModelUVCL(FallThrough[IMDataStructureUVCL]):

    @classmethod
    def holds(cls, obj: IMDataStructureUVCL, parameters: Optional[Dict[str, Any]] = None) -> bool:
        return not EmptyTracesUVCL.holds(obj, parameters)

    @classmethod
    def apply(cls, obj: IMDataStructureUVCL, pool=None, manager=None, parameters: Optional[Dict[str, Any]] = None) -> Optional[
        Tuple[ProcessTree, List[IMDataStructureUVCL]]]:
        log = obj.data_structure
        uvcl_do = UVCL()
        for a in sorted(list(comut.get_alphabet(log))):
            uvcl_do[(a,)] = 1
        uvcl_redo = UVCL()
        im_uvcl_do = IMDataStructureUVCL(uvcl_do)
        im_uvcl_redo = IMDataStructureUVCL(uvcl_redo)
        return ProcessTree(operator=Operator.LOOP), [im_uvcl_do, im_uvcl_redo]


class FlowerModelDFG(FallThrough[IMDataStructureDFG]):
    @classmethod
    def holds(cls, obj: IMDataStructureDFG, parameters: Optional[Dict[str, Any]] = None) -> bool:
        return True

    @classmethod
    def apply(cls, obj: IMDataStructureDFG, pool=None, manager=None, parameters: Optional[Dict[str, Any]] = None) -> Optional[
        Tuple[ProcessTree, List[IMDataStructureDFG]]]:
        activities = set(obj.dfg.start_activities).union(set(obj.dfg.end_activities)).union(set(x[0] for x in obj.dfg.graph)).union(set(x[1] for x in obj.dfg.graph))
        dfg_do = DFG()
        for a in activities:
            dfg_do.start_activities[a] = 1
            dfg_do.end_activities[a] = 1
        dfg_redo = DFG()
        im_dfg_do = IMDataStructureDFG(InductiveDFG(dfg_do))
        im_dfg_redo = IMDataStructureDFG(InductiveDFG(dfg_redo))
        return ProcessTree(operator=Operator.LOOP), [im_dfg_do, im_dfg_redo]




'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from sqlite3 import Connection as SQ3_Connection
from typing import Optional, Dict, Any
from pm4py.util import constants, xes_constants, exec_utils
from enum import Enum


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    TABLE_NAME = "table_name"


def apply(db: SQ3_Connection, parameters: Optional[Dict[Any, Any]] = None) -> str:
    """
    Provides a string containing the required database knowledge for SQLite3 querying
    (in order for the LLM to produce meaningful queries).

    Parameters
    ---------------
    db
        Database
    parameters
        Parameters of the method, including:
        - Parameters.ACTIVITY_KEY => the attribute of the log to be used as activity
        - Parameters.TIMESTAMP_KEY => the attribute of the log to be used as timestamp
        - Parameters.CASE_ID_KEY => the attribute of the log to be used as case identifier
        - Parameters.RESOURCE_KEY => the attribute of the log containing the resource
        - Parameters.TABLE_NAME => the name of the dataframe/table

    Returns
    ---------------
    db_knowledge
        String containing the required database knowledge.
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters, xes_constants.DEFAULT_TIMESTAMP_KEY)
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    resource_key = exec_utils.get_param_value(Parameters.RESOURCE_KEY, parameters, xes_constants.DEFAULT_RESOURCE_KEY)
    table_name = exec_utils.get_param_value(Parameters.TABLE_NAME, parameters, "dataframe")

    descr = """
The underlying database engine is SQLite.
The table is called """+table_name+""".
Each row of the table is corresponding to an event, along with its attributes. There is no separate table containing the process variant.
Please consider the following information: the case identifier is called '"""+case_id_key+"""', the activity is stored inside the attribute '"""+activity_key+"""', the timestamp is stored inside the attribute '"""+timestamp_key+"""', the resource is stored inside the attribute '"""+resource_key+"""'.
If you need to compute timestamp differences, please use the "julianday" function which gets the fractional number of days since noon in Greenwhich on November 24, 4714 B.C.
In SQLite, there is no STDDEV function but you need to compute it as: SQRT(AVG(col*col) - AVG(col)*AVG(col)).
The events are already sorted by timestamp, moreover you cannot use ORDER BY inside a concatenation operator.
    """

    return descr



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.algo.filtering.ocel import ot_endpoints
from pm4py.util import pandas_utils


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Assigns to each event a feature that is 1 when the event completes the lifecycle of at least one object
    of a given type.

    Parameters
    ----------------
    ocel
        OCEL
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    data
        Extracted feature values
    feature_names
        Feature names
    """
    if parameters is None:
        parameters = {}

    ordered_events = parameters["ordered_events"] if "ordered_events" in parameters else ocel.events[ocel.event_id_column].to_numpy()

    object_types = ocel.objects.groupby(ocel.object_type_column)[ocel.object_id_column].agg(list).to_dict()
    endpoints = ocel.relations.groupby(ocel.object_id_column)[ocel.event_id_column].last().to_dict()

    map_endpoints = {ot: set() for ot in object_types}
    for ot in object_types:
        for obj in object_types[ot]:
            for ev in endpoints[obj]:
                map_endpoints[ot].add(ev)

    feature_names = ["@@event_start_"+ot for ot in object_types]
    data = []

    for ev in ordered_events:
        data.append([])
        for ot in object_types:
            data[-1].append(1.0 if ev in map_endpoints[ot] else 0.0)

    return data, feature_names



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.bpmn.obj import BPMN
from typing import Optional, Dict, Any
from copy import deepcopy


def apply(bpmn_graph: BPMN, string_dictio: Dict[str, str], parameters: Optional[Dict[Any, Any]] = None) -> BPMN:
    if parameters is None:
        parameters = {}

    bpmn_graph = deepcopy(bpmn_graph)

    for node in bpmn_graph.get_nodes():
        if isinstance(node, BPMN.Task):
            name = node.get_name()
            if name in string_dictio:
                node.set_name(string_dictio[name])

    return bpmn_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.performance_spectrum.variants import dataframe, log, dataframe_disconnected, log_disconnected
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util import constants, pandas_utils
from typing import Optional, Dict, Any, Union, List
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    PARAMETER_SAMPLE_SIZE = "sample_size"


class Outputs(Enum):
    LIST_ACTIVITIES = "list_activities"
    POINTS = "points"


class Variants(Enum):
    DATAFRAME = dataframe
    LOG = log
    DATAFRAME_DISCONNECTED = dataframe_disconnected
    LOG_DISCONNECTED = log_disconnected


def apply(log: Union[EventLog, EventStream, pd.DataFrame], list_activities: List[str], variant=None, parameters: Optional[Dict[Any, Any]] = None) -> Dict[str, Any]:
    """
    Finds the performance spectrum provided a log/dataframe
    and a list of activities

    Parameters
    -------------
    log
        Event log/Dataframe
    list_activities
        List of activities interesting for the performance spectrum (at least two)
    variant
        Variant to be used (see Variants Enum)
    parameters
        Parameters of the algorithm, including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY

    Returns
    -------------
    ps
        Performance spectrum object (dictionary)
    """
    from pm4py.objects.conversion.log import converter as log_conversion

    if parameters is None:
        parameters = {}

    sample_size = exec_utils.get_param_value(Parameters.PARAMETER_SAMPLE_SIZE, parameters, 10000)

    if len(list_activities) < 2:
        raise Exception("performance spectrum can be applied providing at least two activities!")

    points = None

    if pandas_utils.check_is_pandas_dataframe(log):
        if variant is None:
            variant = Variants.DATAFRAME

        points = exec_utils.get_variant(variant).apply(log, list_activities, sample_size, parameters)

    if points is None:
        if variant is None:
            variant = Variants.LOG

        points = exec_utils.get_variant(variant).apply(log_conversion.apply(log, variant=log_conversion.Variants.TO_EVENT_LOG, parameters=parameters), list_activities, sample_size,
                                                            parameters)

    ps = {Outputs.LIST_ACTIVITIES.value: list_activities, Outputs.POINTS.value: points}

    return ps



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import tempfile
from copy import copy

from graphviz import Digraph
from pm4py.util import exec_utils, constants
from enum import Enum


class Parameters(Enum):
    FORMAT = "format"
    SHOW_LABELS = "show_labels"
    SHOW_NAMES = "show_names"
    FORCE_NAMES = "force_names"
    FILLCOLORS = "fillcolors"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def visualize(ts, parameters=None):
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    show_labels = exec_utils.get_param_value(Parameters.SHOW_LABELS, parameters, True)
    show_names = exec_utils.get_param_value(Parameters.SHOW_NAMES, parameters, True)
    force_names = exec_utils.get_param_value(Parameters.FORCE_NAMES, parameters, None)
    fillcolors = exec_utils.get_param_value(Parameters.FILLCOLORS, parameters, {})
    font_size = exec_utils.get_param_value(Parameters.FONT_SIZE, parameters, 11)
    font_size = str(font_size)
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR)
    enable_graph_title = exec_utils.get_param_value(Parameters.ENABLE_GRAPH_TITLE, parameters, constants.DEFAULT_ENABLE_GRAPH_TITLES)
    graph_title = exec_utils.get_param_value(Parameters.GRAPH_TITLE, parameters, "Transition System")

    for state in ts.states:
        state.label = state.name

    perc_char = '%'

    if force_names:
        nts = copy(ts)
        for index, state in enumerate(nts.states):
            state.name = state.name + " (%.2f)" % (force_names[state])
            state.label = "%.2f" % (force_names[state] * 100.0)
            state.label = state.label + perc_char
        ts = nts

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    filename.close()

    viz = Digraph(ts.name, filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})

    if enable_graph_title:
        viz.attr(label='<<FONT POINT-SIZE="'+str(2*int(font_size))+'">'+graph_title+'</FONT>>', labelloc="top")

    # states
    viz.attr('node')
    for s in ts.states:
        if show_names:
            if s in fillcolors:
                viz.node(str(id(s)), str(s.label), style="filled", fillcolor=fillcolors[s], fontsize=font_size)
            else:
                viz.node(str(id(s)), str(s.label), fontsize=font_size)
        else:
            if s in fillcolors:
                viz.node(str(id(s)), "", style="filled", fillcolor=fillcolors[s], fontsize=font_size)
            else:
                viz.node(str(id(s)), "", fontsize=font_size)
    # arcs
    for t in ts.transitions:
        if show_labels:
            viz.edge(str(id(t.from_state)), str(id(t.to_state)), label=t.name, fontsize=font_size)
        else:
            viz.edge(str(id(t.from_state)), str(id(t.to_state)))

    viz.attr(overlap='false')

    viz.format = image_format

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import copy
import heapq
import importlib.util
from enum import Enum
from typing import List

import pandas as pd

from pm4py.algo.conformance.alignments.process_tree.util import (
    search_graph_pt_replay_semantics as pt_sem,
)
from pm4py.objects.petri_net.utils import align_utils
from pm4py.objects.process_tree.obj import Operator
from pm4py.objects.process_tree.utils import generic as pt_util
from pm4py.util import exec_utils, constants, xes_constants, pandas_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.log.obj import EventLog, Trace
from pm4py.util import typing
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    CORES = "cores"
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    SHOW_PROGRESS_BAR = "show_progress_bar"
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


class SGASearchState:
    def __init__(
        self,
        costs: float,
        index: int,
        state: pt_sem.ProcessTreeState,
        leaves: Optional[List[ProcessTree]] = None,
        parent: Any = None,
        children: List[Any] = None,
    ):
        self.costs = costs  # costs of the solution of this state
        self.index = index  # index 'to be explained'
        self.state = state  # state in the model
        self.leaves = (
            leaves if leaves is not None else list()
        )  # leaves that 'got you here'
        self.parent = parent  # parent search state
        self.children = (
            children if children is not None else set()
        )  # successor search states
        self.path = []

    def __lt__(self, other):
        # Maintain the same sorting logic
        if self.costs < other.costs:
            return True
        elif self.costs == other.costs:
            return self.index > other.index
        else:
            return False

    def __str__(self):
        return (
            "("
            + str(self.costs)
            + ","
            + str(self.index)
            + ","
            + str(self.state)
            + ")"
        )


def _construct_result_dictionary(state: SGASearchState, variant: List[str]):
    result = dict()
    result["cost"] = state.costs
    alignment = list()
    current_state = state
    parent_state = state.parent
    while parent_state is not None:
        if current_state.leaves:
            if current_state.index == len(variant):
                leaf_labels = list(
                    reversed([o.label for o in current_state.leaves])
                )
                if variant[-1] in leaf_labels:
                    i = leaf_labels.index(variant[-1])
                    leaves_reversed = list(reversed(current_state.leaves))
                    for j, leaf in enumerate(leaves_reversed):
                        if i == j:
                            alignment.append((variant[-1], leaf))
                        else:
                            alignment.append((align_utils.SKIP, leaf))
                else:
                    # If we had to insert moves
                    if (
                        current_state.costs - len(current_state.leaves)
                        > parent_state.costs
                    ):
                        alignment.append(
                            (variant[parent_state.index], align_utils.SKIP)
                        )
                    for leaf in reversed(current_state.leaves):
                        alignment.append((align_utils.SKIP, leaf))
            else:
                alignment.append(
                    (variant[parent_state.index], current_state.leaves[-1])
                )
                leaves_reversed = list(reversed(current_state.leaves))[1:]
                for n in leaves_reversed:
                    alignment.append((align_utils.SKIP, n))
        else:
            alignment.append((variant[parent_state.index], align_utils.SKIP))
        parent_state = parent_state.parent
        current_state = current_state.parent
    result["alignment"] = list(reversed(alignment))
    result["optimal"] = True
    result["state"] = state
    return result


def _is_final_tree_state(state, pt):
    return state[(id(pt), pt)] == ProcessTree.OperatorState.CLOSED


def _update_costs_recursive(delta, states):
    for state in states:
        state.costs = state.costs - delta
        _update_costs_recursive(delta, state.children)


def _check_if_state_exists_and_update(
    search_state: SGASearchState, collection
):
    # Still linear, but removing overhead might help a bit.
    match = False
    for alt in collection:
        if alt.index == search_state.index and alt.state == search_state.state:
            match = True
            if search_state.costs < alt.costs:
                _update_costs_recursive(
                    alt.costs - search_state.costs, alt.children
                )
                alt.costs = search_state.costs
                alt.parent = search_state.parent
                alt.leaves = search_state.leaves
    return match


def _add_new_state(state, parent, open, closed):
    if not _check_if_state_exists_and_update(state, closed):
        if not _check_if_state_exists_and_update(state, open):
            parent.children.add(state)
            heapq.heappush(open, state)
    return open, closed


def _obtain_leaves_from_state_path(path, include_tau=False):
    if include_tau:
        return [
            t[0]
            for t in path
            if t[0].operator is None and t[1] == ProcessTree.OperatorState.OPEN
        ]
    else:
        return [
            t[0]
            for t in path
            if t[0].operator is None
            and t[0].label is not None
            and t[1] == ProcessTree.OperatorState.OPEN
        ]


def _need_log_move(old_state, new_state, path) -> bool:
    if _obtain_leaves_from_state_path(path, include_tau=True):
        return True
    choices = [
        o
        for o in old_state.keys()
        if pt_util.is_any_operator_of(o[1], [Operator.XOR, Operator.LOOP])
    ]
    for choice_tuple in choices:
        choice = choice_tuple[1]
        if pt_util.is_operator(choice, Operator.XOR):
            if (
                pt_util.is_in_state(
                    choice, ProcessTree.OperatorState.FUTURE, old_state
                )
                or pt_util.is_in_state(
                    choice, ProcessTree.OperatorState.CLOSED, old_state
                )
            ) and old_state[(id(choice), choice)] != new_state[
                (id(choice), choice)
            ]:
                return True
        elif pt_util.is_operator(choice, Operator.LOOP):
            for loop_child in choice.children:
                if (
                    pt_util.is_in_state(
                        loop_child, ProcessTree.OperatorState.FUTURE, old_state
                    )
                    or pt_util.is_in_state(
                        loop_child, ProcessTree.OperatorState.CLOSED, old_state
                    )
                ) and old_state[(id(loop_child), loop_child)] != new_state[
                    (id(loop_child), loop_child)
                ]:
                    return True
    return False


def align_variant(variant, tree_leaf_set, pt):
    lmcf = [1] * len(variant)
    initial_search_state = SGASearchState(0, 0, pt_sem.get_initial_state(pt))
    closed_set = set()
    open_set = [initial_search_state]
    heapq.heapify(open_set)
    count = 0
    var_len = len(variant)
    while open_set:
        count += 1
        sga_state = heapq.heappop(open_set)
        if (
            _is_final_tree_state(sga_state.state, pt)
            and sga_state.index == var_len
        ):
            return _construct_result_dictionary(sga_state, variant)
        else:
            closed_set.add(sga_state)
            if sga_state.index < len(lmcf):
                current_index = sga_state.index
                current_label = variant[current_index]
                candidates = [
                    t for (i, t) in tree_leaf_set if t.label == current_label
                ]

                need_log_move = len(candidates) == 0
                for leaf in candidates:
                    path, new_state_st = pt_sem.shortest_path_to_enable(
                        leaf, copy.copy(sga_state.state)
                    )
                    if path is None:
                        need_log_move = True
                    else:
                        model_moves = _obtain_leaves_from_state_path(path)
                        # Check if we need a log move due to XOR/LOOP
                        # transitions in the path
                        need_log_move = (
                            need_log_move
                            if need_log_move
                            else _need_log_move(
                                sga_state.state, new_state_st, path
                            )
                        )
                        sync_path, new_state_st = (
                            pt_sem.shortest_path_to_close(leaf, new_state_st)
                        )
                        path.extend(sync_path)
                        leaves = _obtain_leaves_from_state_path(
                            path, include_tau=True
                        )
                        new_state = SGASearchState(
                            sga_state.costs + len(model_moves),
                            current_index + 1,
                            new_state_st,
                            leaves=leaves,
                            parent=sga_state,
                        )
                        new_state.path = tuple(path)
                        open_set, closed_set = _add_new_state(
                            new_state, sga_state, open_set, closed_set
                        )

                if need_log_move:
                    # Add a log move state
                    log_move_state = SGASearchState(
                        sga_state.costs + lmcf[current_index],
                        current_index + 1,
                        copy.copy(sga_state.state),
                        parent=sga_state,
                    )
                    open_set, closed_set = _add_new_state(
                        log_move_state, sga_state, open_set, closed_set
                    )
            else:
                # FINISH
                path, new_state_st = pt_sem.shortest_path_to_close(
                    pt, sga_state.state
                )
                model_moves = _obtain_leaves_from_state_path(
                    path, include_tau=False
                )
                sga_state.state = new_state_st
                sga_state.costs += len(model_moves)
                sga_state.leaves.extend(
                    _obtain_leaves_from_state_path(path, include_tau=True)
                )
                heapq.heappush(open_set, sga_state)
    # Fallback empty result if something went wrong logically (should not
    # happen)
    return _construct_result_dictionary(initial_search_state, variant)


def _apply_variant(variant, tree, leaves, bwc, parameters=None):
    alignment_obj = align_variant(variant, leaves, tree)
    ltrace_bwc = len(variant) + bwc
    alignment_obj["fitness"] = (
        1.0 - alignment_obj["cost"] / ltrace_bwc if ltrace_bwc > 0 else 0.0
    )
    return alignment_obj


def apply_variant(variant, tree, parameters=None):
    leaves = frozenset(pt_util.get_leaves_as_tuples(tree))
    # Compute bwc once
    bwc = align_variant([], leaves, tree)["cost"]
    return _apply_variant(variant, tree, leaves, bwc, parameters)


def _construct_progress_bar(progress_length, parameters):
    if exec_utils.get_param_value(
        Parameters.SHOW_PROGRESS_BAR, parameters, constants.SHOW_PROGRESS_BAR
    ) and importlib.util.find_spec("tqdm"):
        if progress_length > 1:
            from tqdm.auto import tqdm

            return tqdm(
                total=progress_length,
                desc="aligning log, completed variants :: ",
            )
    return None


def _destroy_progress_bar(progress):
    if progress is not None:
        progress.close()
    del progress


def apply_from_variants_list(var_list, tree, parameters=None):
    if parameters is None:
        parameters = {}

    progress = _construct_progress_bar(len(var_list), parameters)
    leaves = frozenset(pt_util.get_leaves_as_tuples(tree))
    # Compute best-worst-cost once
    bwc = align_variant([], leaves, tree)["cost"]
    ret = []
    align_dict = {}
    for variant in var_list:
        if variant not in align_dict:
            align_dict[variant] = _apply_variant(
                variant, tree, leaves, bwc, parameters
            )
            if progress is not None:
                progress.update()
        ret.append(align_dict[variant])
    _destroy_progress_bar(progress)
    return ret


def apply_multiprocessing(
    obj: Union[EventLog, Trace, pd.DataFrame],
    pt: ProcessTree,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Union[typing.AlignmentResult, typing.ListAlignments]:
    if parameters is None:
        parameters = {}

    import multiprocessing

    leaves = frozenset(pt_util.get_leaves_as_tuples(pt))
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    num_cores = exec_utils.get_param_value(
        Parameters.CORES, parameters, multiprocessing.cpu_count() - 2
    )

    if isinstance(obj, Trace):
        variant = tuple(x[activity_key] for x in obj)
        return align_variant(variant, leaves, pt)
    else:
        from concurrent.futures import ProcessPoolExecutor

        # Compute once
        best_worst_cost = align_variant([], leaves, pt)["cost"]

        if pandas_utils.check_is_pandas_dataframe(obj):
            case_id_key = exec_utils.get_param_value(
                Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
            )
            traces = [
                tuple(x)
                for x in obj.groupby(case_id_key)[activity_key]
                .agg(list)
                .to_dict()
                .values()
            ]
        else:
            obj = log_converter.apply(
                obj,
                variant=log_converter.Variants.TO_EVENT_LOG,
                parameters=parameters,
            )
            traces = [tuple(x[activity_key] for x in case) for case in obj]

        with ProcessPoolExecutor(max_workers=num_cores) as executor:
            ret = []
            futures = {}
            align_dict = {}

            for trace in traces:
                if trace not in futures:
                    futures[trace] = executor.submit(
                        align_variant, trace, leaves, pt
                    )

            progress = _construct_progress_bar(len(futures), parameters)
            alignments_ready = 0
            if progress is not None:
                while alignments_ready != len(futures):
                    current = sum(
                        1 for variant in futures if futures[variant].done()
                    )
                    if current > alignments_ready:
                        progress.update(current - alignments_ready)
                    alignments_ready = current

            for variant in traces:
                if variant not in align_dict:
                    al = futures[variant].result()
                    ltrace_bwc = len(variant) + best_worst_cost
                    al["fitness"] = (
                        1.0 - al["cost"] / ltrace_bwc
                        if ltrace_bwc > 0
                        else 0.0
                    )
                    align_dict[variant] = al
                ret.append(align_dict[variant])
            _destroy_progress_bar(progress)
        return ret


def apply(
    obj: Union[EventLog, Trace, pd.DataFrame],
    pt: ProcessTree,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Union[typing.AlignmentResult, typing.ListAlignments]:
    if parameters is None:
        parameters = {}

    leaves = frozenset(pt_util.get_leaves_as_tuples(pt))
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    if isinstance(obj, Trace):
        variant = tuple(x[activity_key] for x in obj)
        return align_variant(variant, leaves, pt)
    else:
        if pandas_utils.check_is_pandas_dataframe(obj):
            case_id_key = exec_utils.get_param_value(
                Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
            )
            traces = [
                tuple(x)
                for x in obj.groupby(case_id_key)[activity_key]
                .agg(list)
                .to_dict()
                .values()
            ]
        else:
            obj = log_converter.apply(
                obj,
                variant=log_converter.Variants.TO_EVENT_LOG,
                parameters=parameters,
            )
            traces = [tuple(x[activity_key] for x in case) for case in obj]

        variants = set(traces)
        bwc = align_variant([], leaves, pt)["cost"]
        align_dict = {}
        ret = []
        progress = _construct_progress_bar(len(variants), parameters)
        for trace in traces:
            if trace not in align_dict:
                align_dict[trace] = _apply_variant(
                    trace, pt, leaves, bwc, parameters
                )
                if progress is not None:
                    progress.update()
            ret.append(align_dict[trace])
        _destroy_progress_bar(progress)
        return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Dict, Union, Tuple
import pandas as pd
from pm4py.util import pandas_utils


def apply(ocel: OCEL) -> Dict[str, Dict[Union[str, Tuple[str, str]], pd.DataFrame]]:
    """
    Gets from an object-centric event log (OCEL)
    a dictionary associating to every event/object/e2o/o2o/change type
    a dataframe containing the associated information.
    This effectively splits the information of different event/object types
    in dense dataframes.

    Running example:

        import pm4py
        from pm4py.objects.ocel.util import ocel_to_dict_types_rel


        ocel = pm4py.read_ocel("tests/input_data/ocel/example_log.jsonocel")
        dct_types_rel = ocel_to_dict_types_rel.apply(ocel)

        # prints the dense dataframes for every event type of the log
        for evt, table in dct_types_rel["ev_types"].items():
            print("\n")
            print(evt)
            print(table)

    Parameters
    -------------
    ocel
        Object-centric event log

    Returns
    -----------
    dct_types_rel
        Dictionary associating to every type the corresponding dense table.

        Keys at the first level:
        - ev_types: pointing to the different event types of the object-centric event log
        - obj_types: pointing to the different object types of the object-centric event log
        - e2o: pointing to the different event-object relationships of the object-centric event log
        - o2o: pointing to the different object-object relationships of the object-centric event log
        - object_changes: pointing to temporal changes in the attributes of the different object types of an OCEL

        Keys at the second level:
        - for "ev_types", "obj_types" and "object_changes": the name of the event/object type related to the dense table
        - for "e2o": a tuple in which the first element is an event type, and the second element is an object type
        - for "o2o": a tuple in which the two elements are interconnected object types

        Value: a Pandas dataframe (dense table).
    """
    ev_types_list = pandas_utils.format_unique(ocel.events[ocel.event_activity].unique())
    obj_types_list = pandas_utils.format_unique(ocel.objects[ocel.object_type_column].unique())
    e2o_list = list(ocel.relations.groupby([ocel.event_activity, ocel.object_type_column]).size().to_dict())

    obj_type_map = ocel.objects[[ocel.object_id_column, ocel.object_type_column]].to_dict("records")
    obj_type_map = {x[ocel.object_id_column]: x[ocel.object_type_column] for x in obj_type_map}
    overall_o2o = ocel.o2o.copy()
    overall_o2o[ocel.object_type_column] = overall_o2o[ocel.object_id_column].map(obj_type_map)
    overall_o2o[ocel.object_type_column+"_2"] = overall_o2o[ocel.object_id_column+"_2"].map(obj_type_map)

    o2o_list = list(overall_o2o.groupby([ocel.object_type_column, ocel.object_type_column+"_2"]).size().to_dict())
    changes_list = pandas_utils.format_unique(ocel.object_changes[ocel.object_type_column].unique())

    dct_types_rel = {"ev_types": {}, "obj_types": {}, "e2o": {}, "o2o": {}, "changes": {}}

    for evt in ev_types_list:
        events = ocel.events[ocel.events[ocel.event_activity] == evt]
        events = events.dropna(axis="columns", how="all")
        dct_types_rel["ev_types"][evt] = events

    for objt in obj_types_list:
        objects = ocel.objects[ocel.objects[ocel.object_type_column] == objt]
        objects = objects.dropna(axis="columns", how="all")
        dct_types_rel["obj_types"][objt] = objects

    for e2ot in e2o_list:
        e2o = ocel.relations[(ocel.relations[ocel.event_activity] == e2ot[0]) & (ocel.relations[ocel.object_type_column] == e2ot[1])]
        e2o = e2o.dropna(axis="columns", how="all")
        dct_types_rel["e2o"][e2ot] = e2o

    for o2ot in o2o_list:
        o2o = overall_o2o[(overall_o2o[ocel.object_type_column] == o2ot[0]) & (overall_o2o[ocel.object_type_column+"_2"] == o2ot[1])]
        o2o = o2o.dropna(axis="columns", how="all")
        dct_types_rel["o2o"][o2ot] = o2o

    for objt in changes_list:
        objects = ocel.object_changes[ocel.object_changes[ocel.object_type_column] == objt]
        objects = objects.dropna(axis="columns", how="all")
        dct_types_rel["changes"][objt] = objects

    return dct_types_rel



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, List, Union
import pandas as pd

from intervaltree import IntervalTree, Interval

from pm4py.objects.log.obj import EventLog
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    EPSILON = "epsilon"
    FILTER_ACTIVITY_COUPLE = "filter_activity_couple"


def log_to_intervals(log: Union[EventLog, pd.DataFrame], parameters: Optional[Dict[Any, Any]] = None) -> List[
    List[Any]]:
    """
    Transforms the event log to a list of intervals that are the
    directly-follows paths in the log (open at the complete timestamp of the source event,
    and closed at the start timestamp of the target event).

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity (default: xes_constants.DEFAULT_NAME_KEY)
        - Parameters.START_TIMESTAMP_KEY => the attribute to be used as start timestamp (default: xes_constants.DEFAULT_TIMESTAMP_KEY)
        - Parameters.TIMESTAMP_KEY => the attribute to be used as completion timestamp (default: xes_constants.DEFAULT_TIMESTAMP_KEY)
        - Parameters.EPSILON => the small gap that is removed from the timestamp of the source event and added to the
            timestamp of the target event to make interval querying possible
        - Parameters.FILTER_ACTIVITY_COUPLE => (optional) keeps only the paths between the specified tuple of two activities.

    Returns
    -----------------
    tree
        Interval tree object (which can be queried at a given timestamp, or range of timestamps)
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters)

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    start_timestamp_key = exec_utils.get_param_value(Parameters.START_TIMESTAMP_KEY, parameters,
                                                     xes_constants.DEFAULT_TIMESTAMP_KEY)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)
    filter_activity_couple = exec_utils.get_param_value(Parameters.FILTER_ACTIVITY_COUPLE, parameters, None)

    ret_list = []

    for trace in log:
        for i in range(len(trace) - 1):
            time_i = trace[i][timestamp_key].timestamp()
            for j in range(i + 1, len(trace)):
                time_j = trace[j][start_timestamp_key].timestamp()
                if time_j >= time_i:
                    if filter_activity_couple is None or (
                            trace[i][activity_key] == filter_activity_couple[0] and trace[j][activity_key] ==
                            filter_activity_couple[1]):
                        ret_list.append([time_i, time_j, trace[i], trace[j], trace.attributes])
                    break

    ret_list.sort(key=lambda x: (x[0], x[1], x[2], x[3]))
    return ret_list


def interval_to_tree(intervals: List[List[Any]], parameters: Optional[Dict[Any, Any]] = None) -> IntervalTree:
    """Internal methods to convert the obtained intervals to the eventual IntervalTree"""
    if parameters is None:
        parameters = {}

    epsilon = exec_utils.get_param_value(Parameters.EPSILON, parameters, 0.00001)

    tree = IntervalTree()

    for inte in intervals:
        tree.add(Interval(inte[0] - epsilon, inte[1] + epsilon,
                          data={"source_event": inte[2], "target_event": inte[3], "trace_attributes": inte[4]}))

    return tree


def apply(log: Union[EventLog, pd.DataFrame], parameters: Optional[Dict[Any, Any]] = None) -> IntervalTree:
    """
    Transforms the event log to an interval tree in which the intervals are the
    directly-follows paths in the log (open at the complete timestamp of the source event,
    and closed at the start timestamp of the target event), and having as associated data the source and the target
    event.

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity (default: xes_constants.DEFAULT_NAME_KEY)
        - Parameters.START_TIMESTAMP_KEY => the attribute to be used as start timestamp (default: xes_constants.DEFAULT_TIMESTAMP_KEY)
        - Parameters.TIMESTAMP_KEY => the attribute to be used as completion timestamp (default: xes_constants.DEFAULT_TIMESTAMP_KEY)
        - Parameters.EPSILON => the small gap that is removed from the timestamp of the source event and added to the
            timestamp of the target event to make interval querying possible
        - Parameters.FILTER_ACTIVITY_COUPLE => (optional) keeps only the paths between the specified tuple of two activities.

    Returns
    -----------------
    tree
        Interval tree object (which can be queried at a given timestamp, or range of timestamps)
    """
    intervals = log_to_intervals(log, parameters=parameters)

    return interval_to_tree(intervals, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import uuid
from typing import Optional, Dict, Any, Set, Tuple
from graphviz import Digraph, Graph
from enum import Enum
from pm4py.util import exec_utils
import tempfile
from pm4py.util import vis_utils, constants
from pm4py.objects.ocel.obj import OCEL


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"
    DIRECTED = "directed"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def ot_to_color(ot: str) -> str:
    ot = int(hash(ot))
    num = []
    while len(num) < 6:
        num.insert(0, ot % 16)
        ot = ot // 16
    ret = "#" + "".join([vis_utils.get_corr_hex(x) for x in num])
    return ret


def apply(ocel: OCEL, graph: Set[Tuple[str, str]], parameters: Optional[Dict[Any, Any]] = None) -> Digraph:
    """
    Visualizes an object graph

    Parameters
    -------------
    ocel
        Object-centric event log
    graph
        Object graph
    parameters
        Variant-specific parameters:
            - Parameters.FORMAT => the format of the visualization ("png", "svg", ...)
            - Parameters.BGCOLOR => the background color
            - Parameters.RANKDIR => the rank direction (LR = left-right, TB = top-bottom)
            - Parameters.DIRECTED => boolean value (draws a directed or undirected graph)

    Returns
    -------------
    gviz
        Graphviz object
    """
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, "transparent")
    rankdir = exec_utils.get_param_value(Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ)
    directed = exec_utils.get_param_value(Parameters.DIRECTED, parameters, True)

    enable_graph_title = exec_utils.get_param_value(Parameters.ENABLE_GRAPH_TITLE, parameters, constants.DEFAULT_ENABLE_GRAPH_TITLES)
    graph_title = exec_utils.get_param_value(Parameters.GRAPH_TITLE, parameters, "Object-Centric Graph")

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    filename.close()

    if directed:
        viz = Digraph("ograph", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})
    else:
        viz = Graph("ograph", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})
    viz.attr('node', shape='ellipse', fixedsize='false')

    if enable_graph_title:
        viz.attr(label='<<FONT POINT-SIZE="20">'+graph_title+'</FONT>>', labelloc="top")
    
    ob_type = ocel.objects.groupby(ocel.object_id_column).first()[ocel.object_type_column].to_dict()

    nodes = set(x[0] for x in graph).union(set(x[1] for x in graph))
    nodes_dict = {}
    for n in nodes:
        v = str(uuid.uuid4())
        nodes_dict[n] = v
        viz.node(v, label=n, fontcolor=ot_to_color(ob_type[n]), color=ot_to_color(ob_type[n]))

    for e in graph:
        viz.edge(nodes_dict[e[0]], nodes_dict[e[1]])

    viz.attr(rankdir=rankdir)
    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any

import pandas as pd

from pm4py.algo.discovery.dfg.adapters.pandas.df_statistics import get_partial_order_dataframe
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.util import typing


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def apply(df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None) -> typing.TemporalProfile:
    """
    Gets the temporal profile from a dataframe.

    Implements the approach described in:
    Stertz, Florian, Jürgen Mangler, and Stefanie Rinderle-Ma. "Temporal Conformance Checking at Runtime based on Time-infused Process Models." arXiv preprint arXiv:2008.07262 (2020).


    Parameters
    ----------
    df
        Dataframe
    parameters
        Parameters, including:
        - Parameters.ACTIVITY_KEY => the column to use as activity
        - Parameters.START_TIMESTAMP_KEY => the column to use as start timestamp
        - Parameters.TIMESTAMP_KEY => the column to use as timestamp
        - Parameters.CASE_ID_KEY => the column to use as case ID

    Returns
    -------
    temporal_profile
        Temporal profile of the dataframe
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)
    start_timestamp_key = exec_utils.get_param_value(Parameters.START_TIMESTAMP_KEY, parameters, None)
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)

    business_hours = exec_utils.get_param_value(Parameters.BUSINESS_HOURS, parameters, False)
    business_hours_slots = exec_utils.get_param_value(Parameters.BUSINESS_HOUR_SLOTS, parameters, constants.DEFAULT_BUSINESS_HOUR_SLOTS)

    workcalendar = exec_utils.get_param_value(Parameters.WORKCALENDAR, parameters, constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR)

    efg = get_partial_order_dataframe(df, activity_key=activity_key, timestamp_key=timestamp_key,
                                      start_timestamp_key=start_timestamp_key, case_id_glue=case_id_key,
                                      keep_first_following=False, business_hours=business_hours, business_hours_slot=business_hours_slots, workcalendar=workcalendar)
    efg = efg[[activity_key, activity_key + "_2", "@@flow_time"]]
    temporal_profile = efg.groupby([activity_key, activity_key + "_2"]).agg(["mean", "std"]).reset_index().fillna(
        0).to_dict("records")

    temporal_profile = {
        (x[(activity_key, "")], x[(activity_key + "_2", "")]): (x[("@@flow_time", "mean")], x[("@@flow_time", "std")])
        for x in temporal_profile}

    return temporal_profile



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import datetime
import os
import tempfile
import uuid
from enum import Enum
from random import randint
from typing import List, Any, Tuple, Dict, Optional, Union

from pm4py.util import exec_utils, constants


class Parameters(Enum):
    FORMAT = "format"
    DOT_SIZE = "dot_size"
    LAYOUT_EXT_MULTIPLIER = "layout_ext_multiplier"
    SHOW_LEGEND = "show_legend"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def __build_unique_values(points_list: List[Any]) -> List[Any]:
    """
    Finds the unique values among the attribute values

    Parameters
    ----------------
    points_list
        Points list

    Returns
    ----------------
    list_unq_values
        List of unique values for each attribute
    """
    unique_values = []
    for i in range(len(points_list[0])):
        unique_values.append(set())
        for j in range(len(points_list)):
            unique_values[-1].add(points_list[j][i])
        unique_values[-1] = sorted(list(unique_values[-1]))
    return unique_values


def __build_corr_dict(unique_values: List[Any]) -> Tuple[List[Any], List[str]]:
    """
    Builds the correspondence between unique values and positions in the graph

    Parameters
    ----------------
    unique_values
        List of unique values for each attribute

    Returns
    ----------------
    corr_dict
        Correspondence between unique values and positions
    attr_type
        A list containing the attribute type for each attribute
    """
    corr_dict = []
    attr_type = []
    for i in range(len(unique_values)):
        corr_dict.append({})
        if isinstance(unique_values[i][0], datetime.datetime):
            min_t = unique_values[i][0].timestamp()
            max_t = unique_values[i][-1].timestamp()
            for idx, v in enumerate(unique_values[i]):
                corr_dict[-1][v] = 1.0 / len(unique_values[i]) + (len(unique_values[i]) - 1) / len(unique_values[i]) * (
                        v.timestamp() - min_t) / (max_t - min_t + 0.00001)
            attr_type.append("date")
        elif isinstance(unique_values[i][0], float) or isinstance(unique_values[i][0], int):
            min_t = unique_values[i][0]
            max_t = unique_values[i][-1]
            for idx, v in enumerate(unique_values[i]):
                corr_dict[-1][v] = 1.0 / len(unique_values[i]) + (len(unique_values[i]) - 1) / len(unique_values[i]) * (
                        v - min_t) / (max_t - min_t + 0.00001)
            attr_type.append("number")
        else:
            for idx, v in enumerate(unique_values[i]):
                corr_dict[-1][v] = float(idx + 1) / float(len(unique_values[i]) + 1)
            attr_type.append("str")
    return corr_dict, attr_type


def __build_color_dict(third_values: List[Any]) -> Dict[Any, str]:
    """
    Builds the color map for the values of the third attribute

    Parameters
    ---------------
    third_values
        Unique values of the third attribute

    Returns
    ---------------
    cmap
        Color map
    """
    color_dict = {}
    for v in third_values:
        color_dict[v] = '#%06X' % randint(0, 0xFFFFFF)
    return color_dict


def apply(points_list: List[Any], attributes: List[str], parameters: Optional[Dict[Union[str, Parameters], Any]] = None):
    """
    Creates the dotted chart with the event stream and the provided attributes

    Parameters
    ---------------
    points_list
        List of points (event stream)
    attributes
        List of attributes that should be included in the dotted chart
    parameters
        Parameters of the visualization, including:
        - Parameters.FORMAT => the format of the visualization (svg, png, ...)
        - Parameters.DOT_SIZE => the size of the dot in the dotted chart

    Returns
    ---------------
    file_path
        Path to the dotted chart visualization
    """
    if parameters is None:
        parameters = {}

    if attributes is None or len(attributes) < 2:
        raise Exception("dotted chart requires the specification of at least two attributes")
    elif len(attributes) > 3:
        raise Exception("dotted chart requires the specification of at most three attributes")

    format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    dot_size = exec_utils.get_param_value(Parameters.DOT_SIZE, parameters, 0.07)
    layout_ext_multiplier = exec_utils.get_param_value(Parameters.LAYOUT_EXT_MULTIPLIER, parameters, 50)
    show_legend = exec_utils.get_param_value(Parameters.SHOW_LEGEND, parameters, True)

    enable_graph_title = exec_utils.get_param_value(Parameters.ENABLE_GRAPH_TITLE, parameters, constants.DEFAULT_ENABLE_GRAPH_TITLES)
    graph_title = exec_utils.get_param_value(Parameters.GRAPH_TITLE, parameters, "Dotted Chart")

    unique_values = __build_unique_values(points_list)
    corr_dict, attr_type = __build_corr_dict(unique_values)
    color_dict = __build_color_dict(unique_values[2]) if len(attributes) == 3 else None

    x_length = 10
    y_length = 10
    if attr_type[0] == "str":
        x_length = max(x_length, len(unique_values[0]) * 1.8)
    if attr_type[1] == "str":
        y_length = max(y_length, len(unique_values[1]) * 0.75)

    x_length *= layout_ext_multiplier
    y_length *= layout_ext_multiplier

    output_file_gv = tempfile.NamedTemporaryFile(suffix=".gv")
    output_file_gv.close()
    output_file_img = tempfile.NamedTemporaryFile(suffix="." + format)
    output_file_img.close()

    lines = ["graph G {"]

    if enable_graph_title:
        lines.append('label=<<FONT POINT-SIZE="20">'+graph_title+'</FONT>>;\nlabelloc="top";\n')

    lines.append("origin [label=\"\", shape=none, width=\"0px\", height=\"0px\", pos=\"0,0!\"];")
    lines.append("rightX [label=\"\", shape=none, width=\"0px\", height=\"0px\", pos=\"%d,0!\"];" % (x_length))
    lines.append("topY [label=\"\", shape=none, width=\"0px\", height=\"0px\", pos=\"0,%d!\"];" % (y_length))
    lines.append("rightXlabel [label=\"%s\", shape=none, width=\"0px\", height=\"0px\", pos=\"%d,0!\"];" % (
        attributes[0], x_length + 1.5))
    lines.append("topYlabel [label=\"%s\", shape=none, width=\"0px\", height=\"0px\", pos=\"0,%d!\"];" % (
        attributes[1], y_length + 1.0))
    lines.append("origin -- rightX [ color=\"black\" ];")
    lines.append("origin -- topY [ color=\"black\" ];")

    if attr_type[0] == "str":
        for k, v in corr_dict[0].items():
            n_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
            lines.append(
                "%s [label=\"%s\", shape=none, width=\"0px\", height=\"0px\", pos=\"%.10f,0!\", fontsize=\"6pt\"];" % (
                    n_id, str(k), v * x_length))

    if attr_type[1] == "str":
        for k, v in corr_dict[1].items():
            n_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
            lines.append(
                "%s [label=\"%s\", shape=none, width=\"0px\", height=\"0px\", pos=\"0,%.10f!\", fontsize=\"6pt\"];" % (
                    n_id, str(k), v * y_length))

    for p in points_list:
        coord_x = corr_dict[0][p[0]]
        coord_y = corr_dict[1][p[1]]
        color = color_dict[p[2]] if color_dict is not None else "blue"
        n_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
        lines.append(
            "%s [label=\"\", shape=circle,  width=\"%.10fpx\", height=\"%.10fpx\", pos=\"%.10f,%.10f!\", fontsize=\"6pt\", style=\"filled\", fillcolor=\"%s\", penwidth=0];" % (
                n_id, dot_size, dot_size, coord_x * x_length, coord_y * y_length, color))

    if color_dict is not None and show_legend:
        lines.append(
            "Legend [label=\"legend (attribute: %s)\", shape=none, width=\"0px\", height=\"0px\", pos=\"0,-%d!\"]" % (
                attributes[2], 1*layout_ext_multiplier))
        row = -1
        for k, v in color_dict.items():
            row -= 1
            n_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
            lines.append(
                "%s [label=\"\", shape=circle, width=\"%.10fpx\", height=\"%.10fpx\", fontsize=\"6pt\", style=\"filled\", fillcolor=\"%s\", pos=\"0,%d!\"]" % (
                    n_id, dot_size, dot_size, v, layout_ext_multiplier*row))
            n_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
            lines.append(
                "%s [label=\"%s\", shape=none, width=\"0px\", height=\"0px\", pos=\"1.5,%d!\", fontsize=\"9pt\"];" % (
                    n_id, str(k), layout_ext_multiplier*row))

    lines.append("}")

    lines = "\n".join(lines)

    F = open(output_file_gv.name, "w")
    F.write(lines)
    F.close()

    os.system("neato -n1 -T" + format + " " + output_file_gv.name + " > " + output_file_img.name)

    return output_file_img.name



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.bpmn.obj import BPMN, Marking


def get_node_by_id(id, bpmn_graph):
    """
    Returns the node with specified id

    Parameters
    ------------
    id
        id string
    bpmn_graph
        BPMN graph

    Returns
    ------------
    node that matches the id
    """
    for node in bpmn_graph.get_nodes():
        if node.get_id() == id:
            return node
    return None


def get_global_start_events(bpmn_graph):
    start_events = []
    for node in bpmn_graph.get_nodes():
        if isinstance(node, BPMN.StartEvent):
            if node.get_process() == bpmn_graph.get_process_id():
                start_events.append(node)
    return start_events


def get_initial_marking(bpmn_graph):
    m = Marking()
    for event in get_global_start_events(bpmn_graph):
        m[event] = 1
    return m


def get_boundary_events_of_activity(activity_id, bpmn_graph):
    boundary_events = []
    for node in bpmn_graph.get_nodes():
        if isinstance(node, BPMN.BoundaryEvent):
            node_activity_id = node.get_activity()
            if node_activity_id == activity_id:
                boundary_events.append(node)
    return boundary_events


def get_external_boundary_events_of_activity(activity_id, bpmn_graph):
    boundary_events = []
    for node in bpmn_graph.get_nodes():
        if isinstance(node, BPMN.MessageBoundaryEvent):
            node_activity_id = node.get_activity()
            if node_activity_id == activity_id:
                boundary_events.append(node)
    return boundary_events


# TODO: Include nodes that are in deeper level, too
def get_all_nodes_inside_process(process_id, bpmn_graph, deep=True):
    nodes = []
    for node in bpmn_graph.get_nodes():
        if deep:
            if process_id in get_processes_deep(node, bpmn_graph):
                nodes.append(node)
        else:
            if node.get_process() == process_id:
                nodes.append(node)
    return nodes


# return the direct children of a subprocess
def get_all_direct_child_subprocesses(process_id, bpmn_graph, include_normal=False):
    nodes = set()
    for node in bpmn_graph.get_nodes():
        if isinstance(node, BPMN.SubProcess) and node.get_process() == process_id:
            if not include_normal:
                boundary_events = get_boundary_events_of_activity(node.get_id(), bpmn_graph)
                # termination events inside subprocess
                termination_events = get_termination_events_of_subprocess_for_pnet(node.get_id(), bpmn_graph)
                if len(boundary_events) > 0 or len(termination_events) > 0:
                    nodes.add(node)
            else:
                nodes.add(node)
    return nodes


# return the direct and indirect children of a subprocess
def get_all_child_subprocesses(process_id, bpmn_graph, include_normal=False):
    sub_processes = set()
    for child in get_all_direct_child_subprocesses(process_id, bpmn_graph, include_normal):
        sub_processes.add(child)
        sub_processes = sub_processes.union(get_all_child_subprocesses(child.get_id(), bpmn_graph, include_normal))
    return sub_processes


def get_processes_deep(node, bpmn_graph):
    if node.get_process() == bpmn_graph.get_process_id():
        return [bpmn_graph.get_process_id()]
    else:
        return [node.get_process()] + get_processes_deep(get_node_by_id(node.get_process(), bpmn_graph), bpmn_graph)


def get_subprocesses_sorted_by_depth(bpmn_graph):
    return sorted([node for node in bpmn_graph.get_nodes() if isinstance(node, BPMN.SubProcess)],
                  key=lambda x: x.get_depth(), reverse=True)


def get_termination_events_of_subprocess(activity_id, bpmn_graph):
    events = []
    for node in bpmn_graph.get_nodes():
        if node.get_process() == activity_id and isinstance(node, BPMN.TerminateEndEvent):
            events.append(node)
    return events


def get_termination_events_of_subprocess_for_pnet(activity_id, bpmn_graph):
    events = []
    for node in bpmn_graph.get_nodes():
        if node.get_process() == activity_id and isinstance(node,
                                                            (BPMN.IntermediateCatchEvent, BPMN.IntermediateThrowEvent)):
            cond = False
            for boundary_event in get_boundary_events_of_activity(activity_id, bpmn_graph):
                if boundary_event.get_name() == node.get_name():
                    cond = True
            if not cond:
                events.append(node)
    return events


def get_start_events_of_subprocess(activity_id, bpmn_graph):
    events = []
    for node in bpmn_graph.get_nodes():
        if node.get_process() == activity_id and isinstance(node, BPMN.StartEvent):
            events.append(node)
    return events


def get_end_events_of_subprocess(activity_id, bpmn_graph):
    events = []
    for node in bpmn_graph.get_nodes():
        if node.get_process() == activity_id and isinstance(node, BPMN.EndEvent):
            events.append(node)
    return events


def bpmn_graph_end_events_as_throw_events(bpmn_graph):
    events = []
    for node in bpmn_graph.get_nodes():
        # subprocess end event
        if node.get_process() != bpmn_graph.get_process_id() and isinstance(node, BPMN.EndEvent) and not isinstance(
                node, BPMN.NormalEndEvent):
            events.append(node)
    return events



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import re

from pm4py.objects import petri_net
from pm4py.objects.transition_system.obj import TransitionSystem
from pm4py.objects.petri_net.utils import align_utils
from pm4py.objects.transition_system import obj as ts
from pm4py.objects.transition_system import utils
from pm4py.util import exec_utils
from enum import Enum
import time


class Parameters(Enum):
    MAX_ELAB_TIME = "max_elab_time"
    PETRI_SEMANTICS = "petri_semantics"


def staterep(name):
    """
    Creates a string representation for a state of a transition system.
    Necessary because graphviz does not support symbols simulation than alphanimerics and '_'.
    TODO: find a better representation.

    Parameters
    ----------
    name: the name of a state

    Returns
    -------
    Version of the name filtered of non-alphanumerical characters (except '_').
    """
    return re.sub(r'\W+', '', name)


def marking_flow_petri(net, im, return_eventually_enabled=False, parameters=None):
    """
    Construct the marking flow of a Petri net

    Parameters
    -----------------
    net
        Petri net
    im
        Initial marking
    return_eventually_enabled
        Return the eventually enabled (visible) transitions
    """
    if parameters is None:
        parameters = {}

    # set a maximum execution time of 1 day (it can be changed by providing the parameter)
    max_exec_time = exec_utils.get_param_value(Parameters.MAX_ELAB_TIME, parameters, 86400)
    semantics = exec_utils.get_param_value(Parameters.PETRI_SEMANTICS, parameters, petri_net.semantics.ClassicSemantics())

    start_time = time.time()

    incoming_transitions = {im: set()}
    outgoing_transitions = {}
    eventually_enabled = {}

    active = [im]
    while active:
        if (time.time() - start_time) >= max_exec_time:
            # interrupt the execution
            return incoming_transitions, outgoing_transitions, eventually_enabled
        m = active.pop()
        enabled_transitions = semantics.enabled_transitions(net, m)
        if return_eventually_enabled:
            eventually_enabled[m] = align_utils.get_visible_transitions_eventually_enabled_by_marking(net, m)
        outgoing_transitions[m] = {}
        for t in enabled_transitions:
            nm = semantics.weak_execute(t, net, m)
            outgoing_transitions[m][t] = nm
            if nm not in incoming_transitions:
                incoming_transitions[nm] = set()
                if nm not in active:
                    active.append(nm)
            incoming_transitions[nm].add(t)

    return incoming_transitions, outgoing_transitions, eventually_enabled


def construct_reachability_graph_from_flow(incoming_transitions, outgoing_transitions,
                                           use_trans_name=False, parameters=None):
    """
    Construct the reachability graph from the marking flow

    Parameters
    ----------------
    incoming_transitions
        Incoming transitions
    outgoing_transitions
        Outgoing transitions
    use_trans_name
        Use the transition name

    Returns
    ----------------
    re_gr
        Transition system that represents the reachability graph of the input Petri net.
    """
    if parameters is None:
        parameters = {}

    re_gr = ts.TransitionSystem()

    map_states = {}
    for s in incoming_transitions:
        map_states[s] = ts.TransitionSystem.State(staterep(repr(s)))
        re_gr.states.add(map_states[s])

    for s1 in outgoing_transitions:
        for t in outgoing_transitions[s1]:
            s2 = outgoing_transitions[s1][t]
            if use_trans_name:
                utils.add_arc_from_to(t.name, map_states[s1], map_states[s2], re_gr)
            else:
                utils.add_arc_from_to(repr(t), map_states[s1], map_states[s2], re_gr)

    return re_gr


def construct_reachability_graph(net, initial_marking, use_trans_name=False, parameters=None) -> TransitionSystem:
    """
    Creates a reachability graph of a certain Petri net.
    DO NOT ATTEMPT WITH AN UNBOUNDED PETRI NET, EVER.

    Parameters
    ----------
    net: Petri net
    initial_marking: initial marking of the Petri net.

    Returns
    -------
    re_gr: Transition system that represents the reachability graph of the input Petri net.
    """
    incoming_transitions, outgoing_transitions, eventually_enabled = marking_flow_petri(net, initial_marking,
                                                                                        parameters=parameters)

    return construct_reachability_graph_from_flow(incoming_transitions, outgoing_transitions,
                                                  use_trans_name=use_trans_name, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Dict, Any, Tuple, Set, List
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel import constants as ocel_constants
from pm4py.objects.ocel.obj import OCEL
from pm4py.statistics.ocel.act_utils import find_associations_from_relations_df


class Parameters(Enum):
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE


def aggregate_events(associations: Dict[str, Dict[str, Set[Tuple[str, str]]]]) -> Dict[
    str, Dict[str, Set[Tuple[str, str]]]]:
    """
    Utility method to calculate the "events" metric from the object-type specific associations.
    """
    ret = {}

    for ot in associations:
        ret[ot] = {}
        for act in associations[ot]:
            ret[ot][act] = set()
            for el in associations[ot][act]:
                ret[ot][act].add(el[0])

    return ret


def aggregate_unique_objects(associations: Dict[str, Dict[str, Set[Tuple[str, str]]]]) -> Dict[
    str, Dict[str, Set[Tuple[str, str]]]]:
    """
    Utility method to calculate the "unique objects" metric from the object-type specific associations.
    """
    ret = {}

    for ot in associations:
        ret[ot] = {}
        for act in associations[ot]:
            ret[ot][act] = set()
            for el in associations[ot][act]:
                ret[ot][act].add(el[1])

    return ret


def aggregate_total_objects(associations: Dict[str, Dict[str, Set[Tuple[str, str]]]]) -> Dict[
    str, Dict[str, Set[Tuple[str, str]]]]:
    """
    Utility method to calculate the "total objects" metric from the object-type specific associations.
    """
    return associations


def find_associations_from_ocel(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> Dict[
    str, Dict[str, List[Tuple[str, str]]]]:
    """
    Associates each object type and activity in the object-centric event log with the combinations
    of event identifiers and objects that are associated to them.

    Parameters
    ------------------
    ocel
        Object-centric event log
    parameters
        Parameters of the method, including:
        - Parameters.EVENT_ID => the attribute to use as event identifier
        - Parameters.OBJECT_ID => the attribute to use as object identifier
        - Parameters.EVENT_ACTIVITY => the attribute to use as activity

    Returns
    -----------------
    dict_associations
        Dictionary that associates each object type (first key) and activity (second key)
        to its (ev. id, obj id.) combinations.
    """
    if parameters is None:
        parameters = {}

    object_type = exec_utils.get_param_value(Parameters.OBJECT_TYPE, parameters, ocel.object_type_column)

    ret = {}

    for ot, relations in ocel.relations.groupby(object_type):
        ret[ot] = find_associations_from_relations_df(relations, parameters=parameters)

    return ret
